@@ -94,7 +94,10 @@ class NaNLabelEncoder(BaseEstimator, TransformerMixin):
             if self.warn:
                 cond = ~np.isin(y, self.classes_)
                 if cond.any():
-                    warnings.warn(f"Found {y[cond].nunique()} unknown classes which were set to NaN", UserWarning)
+                    warnings.warn(
+                        f"Found {np.unique(np.asarray(y)[cond]).size} unknown classes which were set to NaN",
+                        UserWarning,
+                    )
 
             encoded = [self.classes_.get(v, 0) for v in y]
 
@@ -127,6 +130,20 @@ class NaNLabelEncoder(BaseEstimator, TransformerMixin):
         decoded = self.classes_vector_[y]
         return decoded
 
+    def __call__(self, data: (Dict[str, torch.Tensor])) -> torch.Tensor:
+        """
+        Extract prediction from network output. Does not map back to input
+        categories as this would require a numpy tensor without grad-abilities.
+
+        Args:
+            data (Dict[str, torch.Tensor]): Dictionary with entries
+                * prediction: data to de-scale
+
+        Returns:
+            torch.Tensor: prediction
+        """
+        return data["prediction"]
+
 
 class TorchNormalizer(BaseEstimator, TransformerMixin):
     """
