@@ -11,18 +11,69 @@ import warnings
 
 import matplotlib.pyplot as plt
 import numpy as np
+from numpy.lib.type_check import nan_to_num
 import pandas as pd
 from sklearn.exceptions import NotFittedError
 from sklearn.preprocessing import StandardScaler
+from sklearn.utils import shuffle
 from sklearn.utils.validation import check_is_fitted
 import torch
 from torch.distributions import Beta
 from torch.nn.utils import rnn
 from torch.utils.data import DataLoader, Dataset
+from torch.utils.data.sampler import Sampler
 
 from pytorch_forecasting.data.encoders import EncoderNormalizer, GroupNormalizer, NaNLabelEncoder, TorchNormalizer
 
 
+def _find_end_indices(diffs: np.ndarray, max_lengths: np.ndarray, min_length: int) -> Tuple[np.ndarray, np.ndarray]:
+    """
+    Identify end indices in series even if some values are missing.
+
+    Args:
+        diffs (np.ndarray): array of differences to next time step. nans should be filled up with ones
+        max_lengths (np.ndarray): maximum length of sequence by position.
+        min_length (int): minimum length of sequence.
+
+    Returns:
+        Tuple[np.ndarray, np.ndarray]: tuple of arrays where first is end indices and second is list of start
+            and end indices that are currently missing.
+    """
+    missing_start_ends = []
+    end_indices = []
+    length = 1
+    start_idx = 0
+    max_idx = len(diffs) - 1
+    max_length = max_lengths[start_idx]
+
+    for idx, diff in enumerate(diffs):
+        if length >= max_length:
+            while length >= max_length:
+                if length == max_length:
+                    end_indices.append(idx)
+                else:
+                    end_indices.append(idx - 1)
+                length -= diffs[start_idx]
+                if start_idx < max_idx:
+                    start_idx += 1
+                max_length = max_lengths[start_idx]
+        elif length >= min_length:
+            missing_start_ends.append([start_idx, idx])
+        length += diff
+    if len(missing_start_ends) > 0:  # required for numba compliance
+        return np.asarray(end_indices), np.asarray(missing_start_ends)
+    else:
+        return np.asarray(end_indices), np.empty((0, 2), dtype=np.int64)
+
+
+try:
+    import numba
+
+    _find_end_indices = numba.jit(nopython=True)(_find_end_indices)
+except ImportError:
+    pass
+
+
 class TimeSeriesDataSet(Dataset):
     """
     PyTorch Dataset for fitting timeseries models.
@@ -43,9 +94,9 @@ class TimeSeriesDataSet(Dataset):
         self,
         data: pd.DataFrame,
         time_idx: str,
-        target: str,
+        target: Union[str, List[str]],
         group_ids: List[str],
-        weight: Union[str, None] = None,
+        weight: Union[str, List[str], None] = None,
         max_encoder_length: int = 30,
         min_encoder_length: int = None,
         min_prediction_idx: int = None,
@@ -64,7 +115,7 @@ class TimeSeriesDataSet(Dataset):
         add_relative_time_idx: bool = False,
         add_target_scales: bool = False,
         add_encoder_length: Union[bool, str] = "auto",
-        target_normalizer: Union[TorchNormalizer, str] = "auto",
+        target_normalizer: Union[TorchNormalizer, NaNLabelEncoder, str] = "auto",
         categorical_encoders={},
         scalers={},
         randomize_length: Union[None, Tuple[float, float], bool] = False,
@@ -76,9 +127,9 @@ class TimeSeriesDataSet(Dataset):
         Args:
             data: dataframe with sequence data - each row can be identified with ``time_idx`` and the ``group_ids``
             time_idx: integer column denoting the time index
-            target: float column denoting the target
+            target: column denoting the target or list of columns denoting the target - categorical or continous.
             group_ids: list of column names identifying a timeseries
-            weight: column name for weights
+            weight: column name for weights or list of column names corresponding to each target
             max_encoder_length: maximum length to encode
             min_encoder_length: minimum allowed length to encode. Defaults to max_encoder_length.
             min_prediction_idx: minimum time index from where to start predictions
@@ -98,6 +149,8 @@ class TimeSeriesDataSet(Dataset):
                 (e.g. useful for weather categories)
             time_varying_unknown_reals: list of continuous variables that change over
                 time and are not know in the future
+            variable_groups: dictionary mapping a name to a list of columns in the data. The name should be present
+                in a categorical or real class argument, to be able to encode or scale the columns by group.
             dropout_categoricals: list of categorical variables that are unknown when making a forecast without
                 observed history
             constant_fill_strategy: dictionary of column names with constants to fill in missing values if there are
@@ -121,17 +174,30 @@ class TimeSeriesDataSet(Dataset):
         """
         super().__init__()
         self.max_encoder_length = max_encoder_length
-        self.min_encoder_length = min_encoder_length or max_encoder_length
+        assert isinstance(self.max_encoder_length, int), "max encoder length must be integer"
+        if min_encoder_length is None:
+            min_encoder_length = max_encoder_length
+        self.min_encoder_length = min_encoder_length
         assert (
             self.min_encoder_length <= self.max_encoder_length
         ), "max encoder length has to be larger equals min encoder length"
+        assert isinstance(self.min_encoder_length, int), "min encoder length must be integer"
         self.max_prediction_length = max_prediction_length
-        self.min_prediction_length = min_prediction_length or max_prediction_length
+        assert isinstance(self.max_prediction_length, int), "max prediction length must be integer"
+        if min_prediction_length is None:
+            min_prediction_length = max_prediction_length
+        self.min_prediction_length = min_prediction_length
         assert (
             self.min_prediction_length <= self.max_prediction_length
         ), "max prediction length has to be larger equals min prediction length"
-        assert self.min_prediction_length > 0, "prediction length must be larger than 0"
+        assert self.min_prediction_length > 0, "min prediction length must be larger than 0"
+        assert isinstance(self.min_prediction_length, int), "min prediction length must be integer"
         self.target = target
+        if isinstance(target, list) and weight is not None:
+            assert isinstance(weight, list) and len(target) == len(weight), (
+                "if target is a list, then weights should "
+                "be either None or also a list of the same length corresponding to targets"
+            )
         self.weight = weight
         self.time_idx = time_idx
         self.group_ids = [] + group_ids
@@ -151,7 +217,9 @@ class TimeSeriesDataSet(Dataset):
             else:
                 randomize_length = (0.2, 0.05)
         self.randomize_length = randomize_length
-        self.min_prediction_idx = min_prediction_idx or data[self.time_idx].min()
+        if min_prediction_idx is None:
+            min_prediction_idx = data[self.time_idx].min()
+        self.min_prediction_idx = min_prediction_idx
         self.constant_fill_strategy = {} if len(constant_fill_strategy) == 0 else constant_fill_strategy
         self.predict_mode = predict_mode
         self.allow_missings = allow_missings
@@ -173,23 +241,7 @@ class TimeSeriesDataSet(Dataset):
         self.add_encoder_length = add_encoder_length
 
         # target normalizer
-        if isinstance(self.target_normalizer, str) and self.target_normalizer == "auto":
-            coerce_positive = (data[self.target] >= 0).all()
-            if coerce_positive:
-                log_scale = data[self.target].skew() > 2.5
-                coerce_positive = False
-            else:
-                log_scale = False
-            if self.max_encoder_length > 20 and self.min_encoder_length > 1:
-                self.target_normalizer = EncoderNormalizer(coerce_positive=coerce_positive, log_scale=log_scale)
-            else:
-                self.target_normalizer = GroupNormalizer(coerce_positive=coerce_positive, log_scale=log_scale)
-        assert self.min_encoder_length > 1 or not isinstance(
-            self.target_normalizer, EncoderNormalizer
-        ), "EncoderNormalizer is only allowed if min_encoder_length > 1"
-        assert isinstance(
-            self.target_normalizer, TorchNormalizer
-        ), f"target_normalizer has to be either None or of class TorchNormalizer but found {self.target_normalizer}"
+        self._set_target_normalizer(data)
 
         # overwrite values
         self.reset_overwrite_values()
@@ -216,11 +268,11 @@ class TimeSeriesDataSet(Dataset):
         # add decoder length to static real variables
         if self.add_encoder_length:
             assert (
-                "decoder_length" not in data.columns
-            ), "decoder_length is a protected column and must not be present in data"
-            if "decoder_length" not in self.time_varying_known_reals and "decoder_length" not in self.reals:
-                self.static_reals.append("decoder_length")
-            data["decoder_length"] = 0.0  # dummy - real value will be set dynamiclly in __getitem__()
+                "encoder_length" not in data.columns
+            ), "encoder_length is a protected column and must not be present in data"
+            if "encoder_length" not in self.time_varying_known_reals and "encoder_length" not in self.reals:
+                self.static_reals.append("encoder_length")
+            data["encoder_length"] = 0  # dummy - real value will be set dynamiclly in __getitem__()
 
         # validate
         self._validate_data(data)
@@ -234,6 +286,35 @@ class TimeSeriesDataSet(Dataset):
         # convert to torch tensor for high performance data loading later
         self.data = self._data_to_tensors(data)
 
+    def _set_target_normalizer(self, data: pd.DataFrame):
+        """
+        Determine target normalizer.
+
+        Args:
+            data (pd.DataFrame): input data
+        """
+        if isinstance(self.target_normalizer, str) and self.target_normalizer == "auto":
+            if data[self.target].dtype.kind != "f":  # category
+                self.target_normalizer = NaNLabelEncoder()
+                assert not self.add_target_scales, "Target scales can be only added for continous targets"
+            else:
+                coerce_positive = (data[self.target] >= 0).all()
+                if coerce_positive:
+                    log_scale = data[self.target].skew() > 2.5
+                    coerce_positive = False
+                else:
+                    log_scale = False
+                if self.max_encoder_length > 20 and self.min_encoder_length > 1:
+                    self.target_normalizer = EncoderNormalizer(coerce_positive=coerce_positive, log_scale=log_scale)
+                else:
+                    self.target_normalizer = GroupNormalizer(coerce_positive=coerce_positive, log_scale=log_scale)
+        assert self.min_encoder_length > 1 or not isinstance(
+            self.target_normalizer, EncoderNormalizer
+        ), "EncoderNormalizer is only allowed if min_encoder_length > 1"
+        assert isinstance(
+            self.target_normalizer, (TorchNormalizer, NaNLabelEncoder)
+        ), f"target_normalizer has to be either None or of class TorchNormalizer but found {self.target_normalizer}"
+
     def _validate_data(self, data: pd.DataFrame):
         """
         Validate that data will not cause hick-ups later on.
@@ -350,10 +431,12 @@ class TimeSeriesDataSet(Dataset):
                 data[self.target], scales = normalizer.fit_transform(data[self.target], data, return_norm=True)
             elif isinstance(self.target_normalizer, GroupNormalizer):
                 data[self.target], scales = self.target_normalizer.transform(data[self.target], data, return_norm=True)
+            elif isinstance(self.target_normalizer, NaNLabelEncoder):
+                data[self.target] = self.target_normalizer.transform(data[self.target])
             else:
                 data[self.target], scales = self.target_normalizer.transform(data[self.target], return_norm=True)
 
-            # add target sclaes
+            # add target scales
             if self.add_target_scales:
                 for idx, name in enumerate(["center", "scale"]):
                     feature_name = f"{self.target}_{name}"
@@ -462,8 +545,11 @@ class TimeSeriesDataSet(Dataset):
             target_names = "__target__"
         else:
             target_names = ["__target__", "__weight__"]
-        target = torch.tensor(data[target_names].to_numpy(dtype=np.float32), dtype=torch.float)
-        continuous = torch.tensor(data[self.reals].to_numpy(dtype=np.float32), dtype=torch.float)
+        if isinstance(self.target_normalizer, NaNLabelEncoder):
+            target = torch.tensor(data[target_names].to_numpy(dtype=np.long), dtype=torch.long)
+        else:
+            target = torch.tensor(data[target_names].to_numpy(dtype=np.float), dtype=torch.float)
+        continuous = torch.tensor(data[self.reals].to_numpy(dtype=np.float), dtype=torch.float)
 
         tensors = dict(reals=continuous, categoricals=categorical, groups=index, target=target, time=time)
 
@@ -617,12 +703,14 @@ class TimeSeriesDataSet(Dataset):
         df_index["index_start"] = np.arange(len(df_index))
         df_index["time"] = data["__time_idx__"]
         df_index["count"] = (df_index["time_last"] - df_index["time_first"]).astype(int) + 1
-        df_index["group_id"] = g.ngroup()
+        group_ids = g.ngroup()
+        df_index["group_id"] = group_ids
 
-        # calculate maxium index to include from current index_start
-        max_time = (df_index["time"] + self.max_encoder_length + self.max_prediction_length).clip(
-            upper=df_index["count"] + df_index.time_first
-        )
+        min_sequence_length = self.min_prediction_length + self.min_encoder_length
+        max_sequence_length = self.max_prediction_length + self.max_encoder_length
+
+        # calculate maximum index to include from current index_start
+        max_time = (df_index["time"] + max_sequence_length - 1).clip(upper=df_index["count"] + df_index.time_first - 1)
 
         # if there are missing timesteps, we cannot say directly what is the last timestep to include
         # therefore we iterate until it is found
@@ -630,17 +718,20 @@ class TimeSeriesDataSet(Dataset):
             assert (
                 self.allow_missings
             ), "Time difference between steps has been idenfied as larger than 1 - set allow_missings=True"
-            df_index["index_end"] = df_index["index_start"]
-            for _ in range(df_index["count"].max()):
-                new_end_time = (
-                    df_index[["time", "time_diff_to_next"]].iloc[df_index["index_end"]].sum(axis=1).to_numpy()
-                )
-                df_index["index_end"] = df_index["index_end"].where(
-                    new_end_time + 1 > max_time, df_index["index_end"] + 1
-                )
-        else:
-            # direct calculation of end index if there are no missing timesteps in the data
-            df_index["index_end"] = df_index["index_start"] + (max_time - df_index["time"] - 1)
+
+        df_index["index_end"], missing_sequences = _find_end_indices(
+            diffs=df_index.time_diff_to_next.to_numpy(),
+            max_lengths=(max_time - df_index.time).to_numpy() + 1,
+            min_length=min_sequence_length,
+        )
+        # add duplicates but mostly with shorter sequence length for start of timeseries
+        # while the previous steps have ensured that we start a sequence on every time step, the missing_sequences
+        # ensure that there is a sequence that finishes on every timestep
+        if len(missing_sequences) > 0:
+            shortened_sequences = df_index.iloc[missing_sequences[:, 0]].assign(index_end=missing_sequences[:, 1])
+
+            # concatenate shortened sequences
+            df_index = pd.concat([df_index, shortened_sequences], axis=0, ignore_index=True)
 
         # filter out where encode and decode length are not satisfied
         df_index["sequence_length"] = df_index["time"].iloc[df_index["index_end"]].to_numpy() - df_index["time"] + 1
@@ -648,21 +739,32 @@ class TimeSeriesDataSet(Dataset):
         # filter too short sequences
         df_index = df_index[
             # sequence must be at least of minimal prediction length
-            lambda x: (x.sequence_length >= self.min_prediction_length + self.min_encoder_length)
+            lambda x: (x.sequence_length >= min_sequence_length)
             &
             # prediction must be for after minimal prediction index + length of prediction
-            (x["sequence_length"] + x["time"] - 1 >= self.min_prediction_idx - 1 + self.min_prediction_length)
+            (x["sequence_length"] + x["time"] >= self.min_prediction_idx + self.min_prediction_length)
         ]
 
         if predict_mode:  # keep longest element per series (i.e. the first element that spans to the end of the series)
             # filter all elements that are longer than the allowed maximum sequence length
             df_index = df_index[
-                lambda x: (x["time_last"] - x["time"] + 1 <= self.max_prediction_length + self.max_encoder_length)
-                & (x["sequence_length"] >= self.min_prediction_length + self.min_encoder_length)
+                lambda x: (x["time_last"] - x["time"] + 1 <= max_sequence_length)
+                & (x["sequence_length"] >= min_sequence_length)
             ]
             # choose longest sequence
             df_index = df_index.loc[df_index.groupby("group_id").sequence_length.idxmax()]
+
         assert len(df_index) > 0, "filters should not remove entries"
+        # check that all groups/series have at least one entry in the index
+        if not group_ids.isin(df_index.group_id).all():
+            missing_groups = data.loc[~group_ids.isin(df_index.group_id), self.group_ids].drop_duplicates()
+            # decode values
+            for name in missing_groups.columns:
+                missing_groups[name] = self.transform_values(name, missing_groups[name], inverse=True)
+            raise ValueError(
+                f"Min encoder length and/or min prediction length is too large for {len(missing_groups)} series/group. "
+                f"First 10 examples: {list(missing_groups.iloc[:10].to_dict(orient='index').values())}"
+            )
 
         return df_index
 
@@ -683,8 +785,10 @@ class TimeSeriesDataSet(Dataset):
         """
         if betas is None:
             betas = self.randomize_length
-        length = length or self.max_encoder_length
-        min_length = min_length or self.min_encoder_length
+        if length is None:
+            length = self.max_encoder_length
+        if min_length is None:
+            min_length = self.min_encoder_length
         probabilities = Beta(betas[0], betas[1]).sample((1000,))
 
         lengths = ((length - min_length) * probabilities).round() + min_length
@@ -759,11 +863,15 @@ class TimeSeriesDataSet(Dataset):
         time = self.data["time"][index.index_start : index.index_end + 1].clone()
         target = self.data["target"][index.index_start : index.index_end + 1].clone()
         groups = self.data["groups"][index.index_start].clone()
-        target_scale = self.target_normalizer.get_parameters(groups, self.group_ids)
+        if isinstance(self.target_normalizer, NaNLabelEncoder):
+            target_scale = np.zeros(2, dtype=float)
+        else:
+            target_scale = self.target_normalizer.get_parameters(groups, self.group_ids)
 
         # fill in missing values (if not all time indices are specified
         sequence_length = len(time)
         if sequence_length < index.sequence_length:
+            assert self.allow_missings, "allow_missings should be True if sequences have gaps"
             repetitions = torch.cat([time[1:] - time[:-1], torch.ones(1, dtype=time.dtype)])
             indices = torch.repeat_interleave(torch.arange(len(time)), repetitions)
             repetition_indices = torch.cat([torch.tensor([False], dtype=torch.bool), indices[1:] == indices[:-1]])
@@ -850,9 +958,9 @@ class TimeSeriesDataSet(Dataset):
             )
 
         if self.add_encoder_length:
-            data_cont[:, self.reals.index("decoder_length")] = (
-                decoder_length - 0.5 * self.max_encoder_length
-            ) / self.max_encoder_length
+            data_cont[:, self.reals.index("encoder_length")] = (
+                (encoder_length - 0.5 * self.max_encoder_length) / self.max_encoder_length * 2.0
+            )
 
         # rescale target
         if self.target_normalizer is not None and isinstance(self.target_normalizer, EncoderNormalizer):
@@ -949,7 +1057,7 @@ class TimeSeriesDataSet(Dataset):
 
         # target
         target = rnn.pad_sequence([batch[1] for batch in batches], batch_first=True)
-        target_scale = torch.tensor([batch[0]["target_scale"] for batch in batches], dtype=torch.float32)
+        target_scale = torch.tensor([batch[0]["target_scale"] for batch in batches], dtype=torch.float)
 
         return (
             dict(
@@ -968,7 +1076,9 @@ class TimeSeriesDataSet(Dataset):
             target,
         )
 
-    def to_dataloader(self, train: bool = True, batch_size: int = 64, **kwargs) -> DataLoader:
+    def to_dataloader(
+        self, train: bool = True, batch_size: int = 64, batch_sampler: Union[Sampler, str] = None, **kwargs
+    ) -> DataLoader:
         """
         Get dataloader from dataset.
 
@@ -976,6 +1086,11 @@ class TimeSeriesDataSet(Dataset):
             train (bool, optional): if dataloader is used for training or prediction
                 Will shuffle and drop last batch if True. Defaults to True.
             batch_size (int): batch size for training model. Defaults to 64.
+            batch_sampler (Union[Sampler, str]): batch sampler or string. One of
+
+                * "synchronized": ensure that samples in decoder are aligned in time. Does not support missing
+                  values in dataset.
+
             **kwargs: additional arguments to ``DataLoader()``
 
 
@@ -1013,33 +1128,142 @@ class TimeSeriesDataSet(Dataset):
             drop_last=train and len(self) > batch_size,
             collate_fn=self._collate_fn,
             batch_size=batch_size,
+            batch_sampler=batch_sampler,
         )
-
         default_kwargs.update(kwargs)
+        kwargs = default_kwargs
+        if kwargs["batch_sampler"] is not None:
+            sampler = kwargs["batch_sampler"]
+            if isinstance(sampler, str):
+                if sampler == "synchronized":
+                    kwargs["batch_sampler"] = TimeSynchronizedBatchSampler(
+                        self, batch_size=kwargs["batch_size"], shuffle=kwargs["shuffle"], drop_last=kwargs["drop_last"]
+                    )
+                else:
+                    raise ValueError(f"batch_sampler {sampler} unknown - see docstring for valid batch_sampler")
+            del kwargs["batch_size"]
+            del kwargs["shuffle"]
+            del kwargs["drop_last"]
+
         return DataLoader(
             self,
-            **default_kwargs,
+            **kwargs,
         )
 
-    def get_index(self) -> pd.DataFrame:
+    def x_to_index(self, x) -> pd.DataFrame:
         """
-        Data index / order in which items are returned in train=False mode by dataloader.
+        Decode dataframe index from x.
 
         Returns:
             dataframe with time index column for first prediction and group ids
         """
-        decoder_length = pd.DataFrame(
-            dict(
-                prediction_idx=self.data["time"][self.index.index_end.to_numpy()] - (self.min_prediction_idx - 1),
-                sequence_length=self.index.sequence_length,
-                max_prediction_length=self.max_prediction_length,
-            )
-        ).min(axis=1)
-        encoder_lengths = self.index.sequence_length - decoder_length
-        index_data = {self.time_idx: self.index.time + encoder_lengths}
+        index_data = {self.time_idx: x["decoder_time_idx"][:, 0]}
         for id in self.group_ids:
-            index_data[id] = self.data["groups"][:, self.group_ids.index(id)][self.index.index_start.to_numpy()]
+            index_data[id] = x["groups"][:, self.group_ids.index(id)]
             # decode if possible
             index_data[id] = self.transform_values(id, index_data[id], inverse=True)
-        index = pd.DataFrame(index_data, index=self.index.index)
+        index = pd.DataFrame(index_data)
         return index
+
+
+class TimeSynchronizedBatchSampler(Sampler):
+    """
+    Samples mini-batches randomly but in a time-synchronised manner.
+
+    Time-synchornisation means that the time index of the first decoder samples are aligned across the batch.
+    This sampler does not support missing values in the dataset.
+    """
+
+    def __init__(
+        self,
+        data_source: TimeSeriesDataSet,
+        batch_size: int = 64,
+        shuffle: bool = False,
+        drop_last: bool = False,
+    ):
+        """
+        Initialize TimeSynchronizedBatchSampler.
+
+        Args:
+            data_source (TimeSeriesDataSet): timeseries dataset.
+            drop_last (bool): if to drop last mini-batch from a group if it is smaller than batch_size.
+                Defaults to False.
+            shuffle (bool): if to shuffle dataset. Defaults to False.
+            batch_size (int, optional): Number of samples in a mini-batch. This is rather the maximum number
+                of samples. Because mini-batches are grouped by prediction time, chances are that there
+                are multiple where batch size will be smaller than the maximum. Defaults to 64.
+        """
+        # Since collections.abc.Iterable does not check for `__getitem__`, which
+        # is one way for an object to be an iterable, we don't do an `isinstance`
+        # check here.
+        if not isinstance(batch_size, int) or isinstance(batch_size, bool) or batch_size <= 0:
+            raise ValueError(
+                "batch_size should be a positive integer value, " "but got batch_size={}".format(batch_size)
+            )
+        if not isinstance(drop_last, bool):
+            raise ValueError("drop_last should be a boolean value, but got " "drop_last={}".format(drop_last))
+        self.data_source = data_source
+        self.batch_size = batch_size
+        self.drop_last = drop_last
+        self.shuffle = shuffle
+        assert not self.data_source.allow_missings, "allow_missings should be False for time-synchronized mini-batches"
+
+        # construct index from which can be sampled
+        self.construct_batch_groups()
+
+    def construct_batch_groups(self):
+        """
+        Construct index of batches from which can be sampled
+        """
+        index = self.data_source.index
+        # get groups, i.e. group all samples by first predict time
+        decoder_lengths = np.min(
+            [
+                index.time_last - (self.data_source.min_prediction_idx - 1),
+                index.sequence_length - self.data_source.min_encoder_length,
+            ],
+            axis=0,
+        ).clip(max=self.data_source.max_prediction_length)
+        first_prediction_time = index.time + index.sequence_length - decoder_lengths + 1
+        self._groups = pd.RangeIndex(0, len(index.index)).groupby(first_prediction_time)
+
+        # calculate sizes of groups
+        self._group_sizes = {}
+        warns = []
+        for name, group in self._groups.items():  # iterate over groups
+            if self.drop_last:
+                self._group_sizes[name] = len(group) // self.batch_size
+            else:
+                self._group_sizes[name] = (len(group) + self.batch_size - 1) // self.batch_size
+            if self._group_sizes[name] == 0:
+                self._group_sizes[name] = 1
+                warns.append(name)
+        if len(warns) > 0:
+            warnings.warn(
+                f"Less than {self.batch_size} samples available for {len(warns)} prediction times. "
+                f"Use batch size smaller than {self.batch_size}. "
+                f"First 10 prediction times with small batch sizes: {warns[:10]}"
+            )
+        # create index from which can be sampled: index is equal to number of batches
+        # associate index with prediction time
+        self._group_index = np.repeat(list(self._group_sizes.keys()), list(self._group_sizes.values()))
+        # associate index with batch within prediction time group
+        self._sub_group_index = np.concatenate([np.arange(size) for size in self._group_sizes.values()])
+
+    def __iter__(self):
+        if self.shuffle:  # shuffle samples
+            groups = {name: shuffle(group) for name, group in self._groups.items()}
+        else:
+            groups = self._groups
+
+        batch_samples = np.random.permutation(len(self))
+        for idx in batch_samples:
+            name = self._group_index[idx]
+            sub_group = self._sub_group_index[idx]
+            sub_group_start = sub_group * self.batch_size
+            sub_group_end = sub_group_start + self.batch_size
+            batch = groups[name][sub_group_start:sub_group_end]
+            yield batch
+
+    def __len__(self):
+        return len(self._group_index)
