@@ -10,17 +10,17 @@ import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
 from pytorch_lightning import LightningModule
-from pytorch_lightning.metrics.metric import TensorMetric
-from pytorch_lightning.utilities.parsing import AttributeDict, collect_init_args, get_init_args
+from pytorch_lightning.utilities.parsing import get_init_args
 import torch
-from torch import optim, unsqueeze
-from torch.optim.lr_scheduler import LambdaLR, OneCycleLR, ReduceLROnPlateau
+import torch.nn as nn
+from torch.nn.utils import rnn
+from torch.optim.lr_scheduler import LambdaLR, ReduceLROnPlateau
 from torch.utils.data import DataLoader
 from tqdm.notebook import tqdm
 
 from pytorch_forecasting.data import TimeSeriesDataSet
-from pytorch_forecasting.data.encoders import GroupNormalizer
-from pytorch_forecasting.metrics import SMAPE
+from pytorch_forecasting.data.encoders import EncoderNormalizer, GroupNormalizer
+from pytorch_forecasting.metrics import MASE, SMAPE, Metric
 from pytorch_forecasting.optim import Ranger
 from pytorch_forecasting.utils import groupby_apply
 
@@ -56,8 +56,8 @@ class BaseModel(LightningModule):
         log_val_interval: Union[int, float] = None,
         learning_rate: Union[float, List[float]] = 1e-3,
         log_gradient_flow: bool = False,
-        loss: TensorMetric = SMAPE(),
-        logging_metrics: List[TensorMetric] = [],
+        loss: Metric = SMAPE(),
+        logging_metrics: nn.ModuleList = nn.ModuleList([]),
         reduce_on_plateau_patience: int = 1000,
         weight_decay: float = 0.0,
         monotone_constaints: Dict[str, int] = {},
@@ -75,8 +75,9 @@ class BaseModel(LightningModule):
             learning_rate (float, optional): Learning rate. Defaults to 1e-3.
             log_gradient_flow (bool): If to log gradient flow, this takes time and should be only done to diagnose
                 training failures. Defaults to False.
-            loss (TensorMetric, optional): metric to optimize. Defaults to SMAPE().
-            logging_metrics: (List[TensorMetric], optional): list of metrics to log.
+            loss (Metric, optional): metric to optimize. Defaults to SMAPE().
+            logging_metrics (nn.ModuleList[MultiHorizonMetric]): list of metrics that are logged during training.
+                Defaults to [].
             reduce_on_plateau_patience (int): patience after which learning rate is reduced by a factor of 10. Defaults
                 to 1000
             weight_decay (float): weight decay. Defaults to 0.0.
@@ -102,7 +103,7 @@ class BaseModel(LightningModule):
         if not hasattr(self, "loss"):
             self.loss = loss
         if not hasattr(self, "logging_metrics"):
-            self.logging_metrics = logging_metrics
+            self.logging_metrics = nn.ModuleList([l for l in logging_metrics])
         if not hasattr(self, "output_transformer"):
             self.output_transformer = output_transformer
 
@@ -127,25 +128,30 @@ class BaseModel(LightningModule):
         """
         x, y = batch
         log, _ = self.step(x, y, batch_idx, label="train")
+        # log loss
+        self.log("train_loss", log["loss"], on_step=True, on_epoch=True, prog_bar=True)
         return log
 
     def training_epoch_end(self, outputs):
-        log, _ = self.epoch_end(outputs, label="train")
-        return log
+        self.epoch_end(outputs, label="train")
 
     def validation_step(self, batch, batch_idx):
         x, y = batch
-        log, _ = self.step(x, y, batch_idx, label="val")
+        log, _ = self.step(x, y, batch_idx, label="val")  # log loss
+        self.log("val_loss", log["loss"], on_step=False, on_epoch=True, prog_bar=True)
         return log
 
     def validation_epoch_end(self, outputs):
-        log, _ = self.epoch_end(outputs, label="val")
-        return log
+        self.epoch_end(outputs, label="val")
 
     def step(self, x: Dict[str, torch.Tensor], y: torch.Tensor, batch_idx: int, label="train"):
         """
         Run for each train/val step.
         """
+        # pack y sequence if different encoder lengths exist
+        if (x["decoder_lengths"] < x["decoder_lengths"].max()).any():
+            y = rnn.pack_padded_sequence(y, lengths=x["decoder_lengths"], batch_first=True, enforce_sorted=False)
+
         if label == "train" and len(self.hparams.monotone_constaints) > 0:
             # calculate gradient with respect to continous decoder features
             x["decoder_cont"].requires_grad_(True)
@@ -178,26 +184,39 @@ class BaseModel(LightningModule):
             # multiply monotinicity loss by large number to ensure relevance and take to the power of 2
             # for smoothness of loss function
             monotinicity_loss = 10 * torch.pow(monotinicity_loss, 2)
+            if isinstance(self.loss, MASE):
+                loss = self.loss(
+                    prediction, y, encoder_target=x["encoder_target"], encoder_lengths=x["encoder_lengths"]
+                )
+            else:
+                loss = self.loss(prediction, y)
 
-            loss = self.loss(prediction, y) * (1 + monotinicity_loss)
+            loss = loss * (1 + monotinicity_loss)
         else:
             out = self(x)
             out["prediction"] = self.transform_output(out)
 
             # calculate loss
             prediction = out["prediction"]
-            loss = self.loss(prediction, y)
+            if isinstance(self.loss, MASE):
+                loss = self.loss(
+                    prediction, y, encoder_target=x["encoder_target"], encoder_lengths=x["encoder_lengths"]
+                )
+            else:
+                loss = self.loss(prediction, y)
 
-        # log loss
-        tensorboard_logs = {f"{label}_loss": loss}
         # logging losses
         y_hat_detached = prediction.detach()
         y_hat_point_detached = self.loss.to_prediction(y_hat_detached)
         for metric in self.logging_metrics:
-            tensorboard_logs[f"{label}_{metric.name}"] = metric(y_hat_point_detached, y)
-        log = {f"{label}_loss": loss, "log": tensorboard_logs, "n_samples": x["decoder_lengths"].size(0)}
-        if label == "train":
-            log["loss"] = loss
+            if isinstance(metric, MASE):
+                loss_value = metric(
+                    y_hat_point_detached, y, encoder_target=x["encoder_target"], encoder_lengths=x["encoder_lengths"]
+                )
+            else:
+                loss_value = metric(y_hat_point_detached, y)
+            self.log(f"{label}_{metric.name}", loss_value, on_step=label == "train", on_epoch=True)
+        log = {"loss": loss, "n_samples": x["decoder_lengths"].size(0)}
         if self.log_interval(label == "train") > 0:
             self._log_prediction(x, out, batch_idx, label=label)
         return log, out
@@ -216,20 +235,9 @@ class BaseModel(LightningModule):
 
     def epoch_end(self, outputs, label="train"):
         """
-        Run at epoch end for training or validation.
+        Run at epoch end for training or validation. Can be overriden in models.
         """
-        if "callback_metrics" in outputs[0]:  # workaround for pytorch-lightning bug
-            outputs = [out["callback_metrics"] for out in outputs]
-        # log average loss and metrics
-        n_samples = sum([x["n_samples"] for x in outputs])
-        avg_loss = torch.stack([x[f"{label}_loss"] * x["n_samples"] / n_samples for x in outputs]).sum()
-        log_keys = outputs[0]["log"].keys()
-        tensorboard_logs = {
-            f"avg_{key}": torch.stack([x["log"][key] * x["n_samples"] / n_samples for x in outputs]).sum()
-            for key in log_keys
-        }
-
-        return {f"{label}_loss": avg_loss, "log": tensorboard_logs}, outputs
+        pass
 
     def log_interval(self, train: bool):
         """
@@ -268,7 +276,8 @@ class BaseModel(LightningModule):
         x: Dict[str, torch.Tensor],
         out: Dict[str, torch.Tensor],
         idx: int = 0,
-        add_loss_to_title: Union[TensorMetric, bool] = False,
+        add_loss_to_title: Union[Metric, bool] = False,
+        show_future_observed: bool = True,
         ax=None,
     ) -> plt.Figure:
         """
@@ -279,6 +288,7 @@ class BaseModel(LightningModule):
             out: network output
             idx: index of prediction to plot
             add_loss_to_title: if to add loss to title or loss function to calculate. Default to False.
+            show_future_observed: if to show actuals for future. Defaults to True.
             ax: matplotlib axes to plot on
 
         Returns:
@@ -323,8 +333,10 @@ class BaseModel(LightningModule):
             plotter = ax.plot
         else:
             plotter = ax.scatter
+
         # plot observed prediction
-        plotter(x_pred, y[-n_pred:], label=None, c=obs_color)
+        if show_future_observed:
+            plotter(x_pred, y[-n_pred:], label=None, c=obs_color)
 
         # plot prediction
         plotter(x_pred, self.loss.to_prediction(y_hat.unsqueeze(0))[0], label="predicted", c=pred_color)
@@ -350,7 +362,11 @@ class BaseModel(LightningModule):
             else:
                 loss = add_loss_to_title
                 loss.quantiles = self.loss.quantiles
-            ax.set_title(f"Loss {loss(y_hat[None], y[-n_pred:][None]):.3g}")
+            if isinstance(loss, MASE):
+                loss_value = loss(y_hat[None], y[-n_pred:][None], y[:n_pred][None])
+            else:
+                loss_value = loss(y_hat[None], y[-n_pred:][None])
+            ax.set_title(f"Loss {loss_value:.3g}")
         ax.set_xlabel("Time index")
         fig.legend()
         return fig
@@ -532,6 +548,7 @@ class BaseModel(LightningModule):
         output = []
         decode_lenghts = []
         x_list = []
+        index = []
         progress_bar = tqdm(desc="Predict", unit=" batches", total=len(dataloader), disable=not show_progress_bar)
         with torch.no_grad():
             for x, _ in dataloader:
@@ -571,6 +588,8 @@ class BaseModel(LightningModule):
                 output.append(out)
                 if return_x:
                     x_list.append(x)
+                if return_index:
+                    index.append(dataloader.dataset.x_to_index(x))
                 progress_bar.update()
                 if fast_dev_run:
                     break
@@ -594,7 +613,7 @@ class BaseModel(LightningModule):
             x_cat = x_cat
             output.append(x_cat)
         if return_index:
-            output.append(dataloader.dataset.get_index())
+            output.append(pd.concat(index, axis=0, ignore_index=True))
         if return_decoder_lengths:
             output.append(torch.cat(decode_lenghts, dim=0))
         return output
@@ -620,7 +639,7 @@ class BaseModel(LightningModule):
             mode (str, optional): Output mode. Defaults to "dataframe". Either
 
                 * "series": values are average prediction and index are probed values
-                * "dataframe": columns are as obtained by the `dataset.get_index()` method,
+                * "dataframe": columns are as obtained by the `dataset.x_to_index()` method,
                     prediction (which is the mean prediction over the time horizon),
                     normalized_prediction (which are predictions devided by the prediction for the first probed value)
                     the variable name for the probed values
@@ -643,12 +662,17 @@ class BaseModel(LightningModule):
 
         results = []
         progress_bar = tqdm(desc="Predict", unit=" batches", total=len(values), disable=not show_progress_bar)
-        for value in values:
+        for idx, value in enumerate(values):
             # set values
             data.set_overwrite_values(variable=variable, values=value, target=target)
             # predict
             kwargs.setdefault("mode", "prediction")
-            results.append(self.predict(data, **kwargs))
+
+            if idx == 0 and mode == "dataframe":  # need index for returning as dataframe
+                res, index = self.predict(data, return_index=True, **kwargs)
+                results.append(res)
+            else:
+                results.append(self.predict(data, **kwargs))
             # increment progress
             progress_bar.update()
 
@@ -669,9 +693,8 @@ class BaseModel(LightningModule):
             results = results.sum(-1) / (~is_nan).float().sum(-1)
 
             # create dataframe
-            dependencies = data.get_index()
             dependencies = (
-                dependencies.iloc[np.tile(np.arange(len(dependencies)), len(values))]
+                index.iloc[np.tile(np.arange(len(index)), len(values))]
                 .reset_index(drop=True)
                 .assign(prediction=results.flatten())
             )
@@ -885,7 +908,7 @@ class CovariatesMixin:
                 scaler = self.dataset_parameters["scalers"][name]
                 x = np.linspace(-data["std"], data["std"], bins)
                 # reversing normalization for group normalizer is not possible without sample level information
-                if not isinstance(scaler, GroupNormalizer):
+                if not isinstance(scaler, (GroupNormalizer, EncoderNormalizer)):
                     x = scaler.inverse_transform(x)
                     ax.set_xlabel(f"Normalized {name}")
 
@@ -931,3 +954,7 @@ class CovariatesMixin:
             fig.tight_layout()
             fig.legend()
             return fig
+
+
+class AutoRegressiveBaseModel(BaseModel):
+    pass
