@@ -10,15 +10,15 @@ from torch import nn
 from torch.nn.utils import rnn
 
 from pytorch_forecasting.data import TimeSeriesDataSet
-from pytorch_forecasting.metrics import MAE, MAPE, RMSE, SMAPE, MultiHorizonMetric, QuantileLoss
+from pytorch_forecasting.metrics import MAE, MAPE, MASE, RMSE, SMAPE, MultiHorizonMetric, QuantileLoss
 from pytorch_forecasting.models.base_model import BaseModel, CovariatesMixin
+from pytorch_forecasting.models.nn import MultiEmbedding
 from pytorch_forecasting.models.temporal_fusion_transformer.sub_modules import (
     AddNorm,
     GateAddNorm,
     GatedLinearUnit,
     GatedResidualNetwork,
     InterpretableMultiHeadAttention,
-    TimeDistributedEmbeddingBag,
     VariableSelectionNetwork,
 )
 from pytorch_forecasting.utils import autocorrelation, get_embedding_size, integer_histogram
@@ -55,14 +55,16 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         reduce_on_plateau_patience: int = 1000,
         monotone_constaints: Dict[str, int] = {},
         share_single_variable_networks: bool = False,
-        output_transformer: Callable = None,
+        logging_metrics: nn.ModuleList = nn.ModuleList([SMAPE(), MAE(), RMSE(), MAPE()]),
+        **kwargs,
     ):
         """
         Temporal Fusion Transformer for forecasting timeseries - use its :py:meth:`~from_dataset` method if possible.
 
         Implementation of the article
         `Temporal Fusion Transformers for Interpretable Multi-horizon Time Series
-        Forecasting <https://arxiv.org/pdf/1912.09363.pdf>`_.
+        Forecasting <https://arxiv.org/pdf/1912.09363.pdf>`_. The network outperforms DeepAR by Amazon by 36-69%
+        in benchmarks.
 
         Enhancements compared to the original implementation (apart from capabilities added through base model
         such as monotone constraints):
@@ -94,6 +96,9 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
             time_varying_categoricals_decoder: integer of positions of categorical variables for decoder
             time_varying_reals_encoder: integer of positions of continuous variables for encoder
             time_varying_reals_decoder: integer of positions of continuous variables for decoder
+            categorical_groups: dictionary where values
+                are list of categorical variables that are forming together a new categorical
+                variable which is the key in the dictionary
             x_reals: order of continuous variables in tensor passed to forward function
             x_categoricals: order of categorical variables in tensor passed to forward function
             hidden_continuous_size: default for hidden size for processing continous variables (similar to categorical
@@ -118,37 +123,24 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
                 This constraint significantly slows down training. Defaults to {}.
             share_single_variable_networks (bool): if to share the single variable networks between the encoder and
                 decoder. Defaults to False.
+            logging_metrics (nn.ModuleList[MultiHorizonMetric]): list of metrics that are logged during training.
+                Defaults to nn.ModuleList([SMAPE(), MAE(), RMSE(), MAPE()]).
+            **kwargs: additional arguments to :py:class:`~BaseModel`.
         """
         self.save_hyperparameters()
-        super().__init__()
         # store loss function separately as it is a module
         assert isinstance(loss, MultiHorizonMetric), "Loss has to of class `MultiHorizonMetric`"
-        self.loss = loss
-        self.output_transformer = output_transformer
-        self.logging_metrics = [SMAPE(), MAE(), RMSE(), MAPE()]
+        super().__init__(loss=loss, logging_metrics=logging_metrics, **kwargs)
 
         # processing inputs
         # embeddings
-        self.input_embeddings = nn.ModuleDict()
-        for name in self.hparams.embedding_sizes.keys():
-            embedding_size = min(self.hparams.embedding_sizes[name][1], self.hparams.hidden_size)
-            # convert to list to become mutable
-            self.hparams.embedding_sizes[name] = list(self.hparams.embedding_sizes[name])
-            self.hparams.embedding_sizes[name][1] = embedding_size
-            if name in self.hparams.categorical_groups:  # embedding bag if related embeddings
-                self.input_embeddings[name] = TimeDistributedEmbeddingBag(
-                    self.hparams.embedding_sizes[name][0], embedding_size, mode="sum", batch_first=True
-                )
-            else:
-                if name in self.hparams.embedding_paddings:
-                    padding_idx = 0
-                else:
-                    padding_idx = None
-                self.input_embeddings[name] = nn.Embedding(
-                    self.hparams.embedding_sizes[name][0],
-                    embedding_size,
-                    padding_idx=padding_idx,
-                )
+        self.input_embeddings = MultiEmbedding(
+            embedding_sizes=self.hparams.embedding_sizes,
+            categorical_groups=self.hparams.categorical_groups,
+            embedding_paddings=self.hparams.embedding_paddings,
+            x_categoricals=self.hparams.x_categoricals,
+            max_embedding_size=self.hparams.hidden_size,
+        )
 
         # continuous variable processing
         self.prescalers = nn.ModuleDict(
@@ -197,7 +189,6 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         )
 
         # create single variable grns that are shared across decoder and encoder
-
         if self.hparams.share_single_variable_networks:
             self.shared_single_variable_grns = nn.ModuleDict()
             for name, input_size in encoder_input_sizes.items():
@@ -420,7 +411,7 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         #   matter in the future than the past)
         #   or alternatively using the same layer but allowing forward attention - i.e. only masking out non-available
         #   data and self
-        decoder_mask = attend_step > predict_step
+        decoder_mask = attend_step >= predict_step
         # do not attend to steps where data is padded
         encoder_mask = self._get_mask(encoder_lengths.max(), encoder_lengths)
         # combine masks along attended time - first encoder and then decoder
@@ -443,20 +434,7 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         x_cont = torch.cat([x["encoder_cont"], x["decoder_cont"]], dim=1)  # concatenate in time dimension
         timesteps = x_cont.size(1)  # encode + decode length
         max_encoder_length = int(encoder_lengths.max())
-        input_vectors = {}
-        for name, emb in self.input_embeddings.items():
-            if name in self.hparams.categorical_groups:
-                input_vectors[name] = emb(
-                    x_cat[
-                        ...,
-                        [
-                            self.hparams.x_categoricals.index(cat_name)
-                            for cat_name in self.hparams.categorical_groups[name]
-                        ],
-                    ]
-                )
-            else:
-                input_vectors[name] = emb(x_cat[..., self.hparams.x_categoricals.index(name)])
+        input_vectors = self.input_embeddings(x_cat)
         input_vectors.update({name: x_cont[..., idx].unsqueeze(-1) for idx, name in enumerate(self.hparams.x_reals)})
 
         # Embedding and variable selection
@@ -589,7 +567,6 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         run at each step for training or validation
         """
         # extract data and run model
-        y = rnn.pack_padded_sequence(y, lengths=x["decoder_lengths"], batch_first=True, enforce_sorted=False)
         log, out = super().step(x, y, batch_idx, label=label)
         # calculate interpretations etc for latter logging
         if self.log_interval(label == "train") > 0:
@@ -606,10 +583,8 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         """
         run at epoch end for training or validation
         """
-        log, out = super().epoch_end(outputs, label=label)
         if self.log_interval(label == "train") > 0:
-            self._log_interpretation(out, label=label)
-        return log, out
+            self._log_interpretation(outputs, label=label)
 
     def interpret_output(
         self,
@@ -659,7 +634,7 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         # attention is batch x time x heads x time_to_attend
         # average over heads + only keep prediction attention and attention on observed timesteps
         attention = out["attention"][
-            :, attention_prediction_horizon, :, : out["encoder_lengths"].max() + attention_prediction_horizon + 1
+            :, attention_prediction_horizon, :, : out["encoder_lengths"].max() + attention_prediction_horizon
         ].mean(1)
 
         if reduction != "none":  # if to average over batches
@@ -671,14 +646,12 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
             for i in range(len(attention)):  # very inefficient but does the trick
                 if 0 < out["encoder_lengths"][i] < attention.size(1) - attention_prediction_horizon - 1:
                     relevant_attention = attention[
-                        i, : out["encoder_lengths"][i] + attention_prediction_horizon + 1
+                        i, : out["encoder_lengths"][i] + attention_prediction_horizon
                     ].clone()
                     if attention_as_autocorrelation:
                         relevant_attention = autocorrelation(relevant_attention)
-                    attention[i, -out["encoder_lengths"][i] - attention_prediction_horizon - 1 :] = relevant_attention
-                    attention[
-                        i, : attention.size(1) - out["encoder_lengths"][i] - attention_prediction_horizon - 1
-                    ] = 0.0
+                    attention[i, -out["encoder_lengths"][i] - attention_prediction_horizon :] = relevant_attention
+                    attention[i, : attention.size(1) - out["encoder_lengths"][i] - attention_prediction_horizon] = 0.0
                 elif attention_as_autocorrelation:
                     attention[i] = autocorrelation(attention[i])
 
@@ -692,12 +665,12 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
                 raise ValueError(f"Unknown reduction {reduction}")
 
             attention = torch.zeros(
-                self.hparams.max_encoder_length + attention_prediction_horizon + 1, device=self.device
+                self.hparams.max_encoder_length + attention_prediction_horizon, device=self.device
             ).scatter(
                 dim=0,
                 index=torch.arange(
-                    self.hparams.max_encoder_length + attention_prediction_horizon + 1 - attention.size(-1),
-                    self.hparams.max_encoder_length + attention_prediction_horizon + 1,
+                    self.hparams.max_encoder_length + attention_prediction_horizon - attention.size(-1),
+                    self.hparams.max_encoder_length + attention_prediction_horizon,
                     device=self.device,
                 ),
                 src=attention,
@@ -722,6 +695,7 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         idx: int,
         plot_attention: bool = True,
         add_loss_to_title: bool = False,
+        show_future_observed: bool = True,
         ax=None,
     ) -> plt.Figure:
         """
@@ -733,6 +707,7 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
             idx (int): sample index
             plot_attention: if to plot attention on secondary axis
             add_loss_to_title: if to add loss to title. Default to False.
+            show_future_observed: if to show actuals for future. Defaults to True.
             ax: matplotlib axes to plot on
 
         Returns:
@@ -740,7 +715,9 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
         """
 
         # plot prediction as normal
-        fig = super().plot_prediction(x, out, idx=idx, add_loss_to_title=add_loss_to_title, ax=ax)
+        fig = super().plot_prediction(
+            x, out, idx=idx, add_loss_to_title=add_loss_to_title, show_future_observed=show_future_observed, ax=ax
+        )
 
         # add attention on secondary axis
         if plot_attention:
@@ -748,11 +725,10 @@ class TemporalFusionTransformer(BaseModel, CovariatesMixin):
             ax = fig.axes[0]
             ax2 = ax.twinx()
             ax2.set_ylabel("Attention")
-            max_encoder_length = out["encoder_lengths"].max()
             encoder_length = x["encoder_lengths"][idx]
             ax2.plot(
-                torch.arange(-encoder_length, 1),
-                interpretation["attention"][idx, list(range(encoder_length)) + [max_encoder_length]].detach().cpu(),
+                torch.arange(-encoder_length, 0),
+                interpretation["attention"][idx, :encoder_length].detach().cpu(),
                 alpha=0.2,
                 color="k",
             )
