@@ -9,7 +9,7 @@ import optuna
 from optuna.integration import PyTorchLightningPruningCallback, TensorBoardCallback
 import pytorch_lightning as pl
 from pytorch_lightning import Callback
-from pytorch_lightning.callbacks import LearningRateLogger
+from pytorch_lightning.callbacks import LearningRateMonitor
 from pytorch_lightning.loggers import TensorBoardLogger
 import statsmodels.api as sm
 import torch
@@ -46,6 +46,7 @@ def optimize_hyperparameters(
     use_learning_rate_finder: bool = True,
     trainer_kwargs: Dict[str, Any] = {},
     log_dir: str = "lightning_logs",
+    study: optuna.Study = None,
     **kwargs,
 ) -> optuna.Study:
     """
@@ -77,6 +78,7 @@ def optimize_hyperparameters(
             `PyTorch Lightning trainer <https://pytorch-lightning.readthedocs.io/en/latest/trainer.html>`_ such
             as ``limit_train_batches``. Defaults to {}.
         log_dir (str, optional): Folder into which to log results for tensorboard. Defaults to "lightning_logs".
+        study (optuna.Study, optional): study to resume. Will create new study by default.
         **kwargs: Additional arguments for the :py:class:`~TemporalFusionTransformer`.
 
     Returns:
@@ -97,7 +99,7 @@ def optimize_hyperparameters(
         # TensorBoard. We don't use any logger here as it requires us to implement several abstract
         # methods. Instead we setup a simple callback, that saves metrics from each validation step.
         metrics_callback = MetricsCallback()
-        learning_rate_callback = LearningRateLogger()
+        learning_rate_callback = LearningRateMonitor()
         logger = TensorBoardLogger(log_dir, name="optuna", version=trial.number)
         gradient_clip_val = trial.suggest_loguniform("gradient_clip_val", *gradient_clip_val_range)
         trainer = pl.Trainer(
@@ -105,8 +107,11 @@ def optimize_hyperparameters(
             max_epochs=max_epochs,
             gradient_clip_val=gradient_clip_val,
             gpus=[0] if torch.cuda.is_available() else None,
-            callbacks=[metrics_callback, learning_rate_callback],
-            early_stop_callback=PyTorchLightningPruningCallback(trial, monitor="val_loss"),
+            callbacks=[
+                metrics_callback,
+                learning_rate_callback,
+                PyTorchLightningPruningCallback(trial, monitor="val_loss"),
+            ],
             logger=logger,
             **trainer_kwargs,
         )
@@ -134,7 +139,7 @@ def optimize_hyperparameters(
                 gpus=[0] if torch.cuda.is_available() else None,
                 logger=False,
             )
-            res = lr_trainer.lr_find(
+            res = lr_trainer.tuner.lr_find(
                 model,
                 train_dataloader=train_dataloader,
                 val_dataloaders=val_dataloader,
@@ -165,6 +170,7 @@ def optimize_hyperparameters(
 
     # setup optuna and run
     pruner = optuna.pruners.SuccessiveHalvingPruner()
-    study = optuna.create_study(direction="minimize", pruner=pruner)
+    if study is None:
+        study = optuna.create_study(direction="minimize", pruner=pruner)
     study.optimize(objective, n_trials=n_trials, timeout=timeout)
     return study
